#include "framecpp/config.h"

#include <assert.h>

#include <stdexcept>
#include <iostream>

#include "general/AtExit.hh"
#include "general/gpstime.hh"
#include "general/mutexlock.hh"

#include "framecpp/Common/Description.hh"
#include "framecpp/Common/FrameSpec.hh"

#include "framecpp/Common/IOStream.hh"

#include "framecpp/Common/FrDetector.hh"
#include "framecpp/Common/FrStatData.hh"

#include "framecpp/Common/FrameSpec.tcc"

using General::GPSTime;

namespace FrameCPP
{
  namespace Common
  {
    FrameSpec::frame_spec_container::
    frame_spec_container( )
    {
    }

    FrameSpec::frame_spec_container::
    ~frame_spec_container( )
    {
      for ( const_iterator
	      cur = begin( ),
	      last = end( );
	    cur != last;
	    ++cur )
      {
	delete cur->second;
      }
      erase( begin( ), end( ) );
    }

    FrameSpec::ObjectInterface::
    ~ObjectInterface( )
    {
    }

    FrameSpec::Object::
    Object( class_type Class, const Description* Description )
      : m_class( Class ),
	m_desc( Description )
    {
    }

    FrameSpec::Object::
    Object( const Object& Source )
      : m_class( Source.m_class ),
	m_desc( Source.m_desc )
    {
    }

    FrameSpec::Object::
    ~Object( )
    {
    }

    OStream& FrameSpec::Object::
    WriteNext( OStream& Stream ) const
    {
      Stream.NextPtrStruct( this ).Write( Stream );
      return Stream;
    }

    FrameSpec::class_type FrameSpec::Object::
    getClass( ) const
    {
      return GetClass( );
    }

    FrameSpec::Object* FrameSpec::ObjectInterface::
    Create( ) const
    {
      std::ostringstream	msg;
      const Description* 	d( GetDescription( ) );

      msg << "Cannot create object of class: " << getClass( );
      if ( d )
      {
	msg << " (" << d->GetName( ) << ")";
      }


      throw std::runtime_error( msg.str( ) );
    }

    FrameSpec::Object* FrameSpec::ObjectInterface::
    Clone( ) const
    {
      const Description* d( GetDescription( ) );
      std::ostringstream	msg;

      msg << "Cannot clone object of class: " << getClass( );
      if ( d )
      {
	msg << " (" << d->GetName( ) << ")";
      }

      throw std::runtime_error( msg.str( ) );
    }

    FrameSpec::Object* FrameSpec::ObjectInterface::
    DemoteObject(  INT_2U Target, Object* Obj, Common::IStream* Stream ) 
    {
      std::auto_ptr< Object >	retval;
      FrameSpec::Object*	previous = retval.get( );
      FrameSpec::Object*	current = Obj;

      do
      {
	previous = retval.get( );
	retval = current->demote( Target, Stream );
	current = retval.get( );
      } while ( ( current ) && ( current != previous ) );
      return retval.release( );
    }

    FrameSpec::Object* FrameSpec::ObjectInterface::
    PromoteObject( INT_2U Target, INT_2U Source,
		   Object* Obj, Common::IStream* Stream )
    {
      std::auto_ptr< Object >	retval;
      if ( Obj == (Object*)NULL )
      {
	return retval.release( );
      }

      //-----------------------------------------------------------------
      // Obtain information about the targeted framespec
      //-----------------------------------------------------------------
      const FrameSpec::Info* info( FrameSpec::SpecInfo( Target ) );

      if ( info )
      {
	const Object*
	  target( info->FrameObject( FrameSpec::Info::
				     frame_object_types( Obj->GetClass( ) ) ) );
	if ( target )
	{
	  //-------------------------------------------------------------
	  // Ask the highest object to upconvert. This is recursive so
	  //  earlier versions do not need to know about later versions.
	  //-------------------------------------------------------------
	  retval = target->promote( Source, Obj, Stream );
	}
      }
      return retval.release( );
    }

    void FrameSpec::ObjectInterface::
    VerifyObject( Verify& Verifier, IFrameStream& Stream ) const
    {
    }

    FrameSpec::Info::
    Info( version_type MajorVersion, version_type MinorVersion,
	  verification_func_type VerificationFunc )
      : m_version( MajorVersion ),
	m_version_minor( MinorVersion ),
	m_verification_func( VerificationFunc )
    {
    }

    FrameSpec::Info::
    ~Info( )
    {
      for ( object_container::const_iterator
	      cur = m_definitions.begin( ),
	      last = m_definitions.end( );
	    cur != last;
	    ++cur )
      {
	delete cur->second;
      }
      m_definitions.clear( );
    }

    const FrameSpec::Object* FrameSpec::Info::
    FrameObject( frame_object_types ObjectId ) const
    {
      object_container::const_iterator retval
	= m_definitions.find( ObjectId );
      if ( retval == m_definitions.end( ) )
      {
	std::ostringstream	msg;

	msg << "Unable to locate Object: " << ObjectId
	    << " for Version: " << Version( )
	    << " of the frame specification"
	  ;
	throw std::range_error( msg.str( ) );
      }
      return retval->second;
    }

    void FrameSpec::Info::
    FrameObject( const Object* ObjectTemplate )
    {
      m_definitions[ ObjectTemplate->GetClass( ) ] = ObjectTemplate;
    }
    
    inline FrameSpec::frame_spec_container& FrameSpec::
    frame_specs( )
    {
      static frame_spec_container m_frame_specs;

      return m_frame_specs;
    }

    FrameSpec::Info* FrameSpec::
    SpecInfo( version_type Version )
    {
      frame_spec_container::const_iterator retval
	= frame_specs( ).find( Version );
      
      if ( retval == frame_specs( ).end( ) )
      {
	std::ostringstream	msg;

	msg << "No frame definition found for version " << Version 
	    << " of the frame specification"
	    << std::endl;
	throw std::range_error( msg.str( ) );
      }
      return retval->second;
    }

    void FrameSpec::
    SpecInfo( version_type Version, FrameSpec::Info* I )
    {
      if ( I )
      {
	static MutexLock::lock_type	key = MutexLock::Initialize( );

	MutexLock		lock( key );

	frame_specs( )[ Version ] = I;
      }
    }

    //-------------------------------------------------------------------
    //-------------------------------------------------------------------
    template<> FrameSpec::size_type
    Bytes< GPSTime >( const GPSTime& Source )
    {
      return
	sizeof( INT_4U )
	+ sizeof( INT_4U )
	;
    }

#define INSTANTIATE( LM_TYPE, LM_CRC_TYPE )	\
    template FrameSpec::ObjectWithChecksum< LM_TYPE, LM_CRC_TYPE >

    INSTANTIATE( FrameSpec::Object, INT_4U )::~ObjectWithChecksum( );
    INSTANTIATE( FrDetector, INT_4U )::~ObjectWithChecksum( );
    INSTANTIATE( FrStatData, INT_4U )::~ObjectWithChecksum( );

#undef INSTANTIATE

  } // namespace - Common

} // namespace - FrameCPP
